//
//  ViewController.swift
//  grdb1
//

import UIKit
import GRDB
import Async


class Person : Record {
    var id: Int64?
    var firstName: String?
    var lastName: String?
    
    var fullName: String {
        return [firstName, lastName].flatMap { $0 }.joined(separator: " ")
    }
    
    init(firstName: String?, lastName: String?) {
        self.id = nil
        self.firstName = firstName
        self.lastName = lastName
        super.init()
    }
    
    
    //: Subclasses of Record have to override the methods that define how they interact with the database.
    //:
    //: 1. The table name:
    
    override class var databaseTableName: String {
        return "persons"
    }
    
    //: 2. How to build a Person from a database row:
    
    required init(row: Row) {
        id = row["id"]
        firstName = row["firstName"]
        lastName = row["lastName"]
        super.init(row: row)
    }
    
    //: 3. The dictionary of values that are stored in the database:
    
    override func encode(to container: inout PersistenceContainer) {
        container["id"] = id
        container["firstName"] = firstName
        container["lastName"] = lastName
    }
    
    //: 4. When relevant, update the person's id after a database row has been inserted:
    
    override func didInsert(with rowID: Int64, for column: String?) {
        id = rowID
    }
}

class ViewController: UIViewController {

    var dbQueue: DatabasePool!
    private var trackController: FetchedRecordsController<Person>!
    
    override func viewDidLoad() {
        super.viewDidLoad()
        
        self.initdb()
        
        self.initTracker()
        
        self.test()
        
        Async.main(after: 5) {
            print("Second try...")
            self.test()
        }
        
    }

    func initdb() {
        let libraryPath = NSSearchPathForDirectoriesInDomains(.libraryDirectory, .userDomainMask, true).first! as NSString
        let databasePath = libraryPath.appendingPathComponent("db.sqlite")
        print("DB File: \(databasePath)")
        
        var configuration = Configuration()
        //configuration.trace = { print($0) }
        self.dbQueue = try! DatabasePool(path: databasePath, configuration: configuration)
        
        try? self.dbQueue.write { db in
            try db.execute(
                "CREATE TABLE persons (" +
                    "id INTEGER PRIMARY KEY, " +
                    "firstName TEXT, " +
                    "lastName TEXT" +
                ")")
        }
    }
    
    func initTracker() {
        self.trackController = try! FetchedRecordsController<Person>(
            self.dbQueue,
            sql: "SELECT * FROM persons")
        
        self.trackController.trackChanges(willChange: { controller in
            print("************ willChange !!!")
        }, onChange: { (controller, mediaItem, change) in
            print("onChange: \(mediaItem) - change: \(change)")
        }, didChange: { controller in
            print("************ didChange !!!")
        })
        self.trackController.trackErrors { (controller, error) in
            print("Missed a transaction because \(error)")
        }
        try! self.trackController.performFetch()
        
    }
    
    func test() {
        try! self.dbQueue.writeInTransaction { db in
            for _ in 1...1000 {
                let arthur = Person(firstName: "Arthur", lastName: "Miller")
                let barbra = Person(firstName: "Barbra", lastName: "Streisand")
                let cinderella = Person(firstName: "Cinderella", lastName: nil)
                
                try arthur.insert(db)
                try barbra.insert(db)
                try cinderella.insert(db)
            }
            return .commit
        }
        
        
    }
    
}

